# SnapMatcher.py
# Copyright 2005, 2006, 2007 Matt Heinzen

# This file is part of SnapMatcher 0.2.
#
# SnapMatcher is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# SnapMatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with File Duplicate Finder; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import Image
from scipy import *
import os, sys, string, re
from math import *

class ImageRecord:
	"""Stores the path, file size and signature vector for a single image."""
	def __init__(self, path, fileSize, coefs4x4):
		self.path = path
		self.fileSize = fileSize
		self.coefs4x4 = coefs4x4


def calcImageSig(img, steps):
	"""Calculates a signature vector for the given image and grid resolution."""
	return list(img.resize((steps, steps), Image.ANTIALIAS).getdata())


def createImageDB(rootDir, dirList, fileExts = [], mode = 1):
	"""
	Returns a dictionary of image file paths to ImageRecords for all image files
	found in the given directories.

	Only files matching one of the provided file extensions will be processed if
	a list of extensions is provided.
	"""

	# build a regular expression to determine if a filename matches any of the
	# provided file extensions, or any file if no extensions are provided
	# the pattern will apply to lowercase filenames only
	if len(fileExts) == 0:
		extRE = re.compile("")
	else:
		pattern = "\.("
		for ext in fileExts:
			pattern = pattern + ext + "|"
		pattern = pattern[0:-1] + ")$"
		extRE = re.compile(pattern)

	# build a regex to detect hidden directories
	pattern = os.sep + "\."
	hiddenRE = re.compile(pattern)

	imageDB = []

	for directory in dirList:
		for root, dirs, files in os.walk(directory):
			if re.search(hiddenRE, root) == None:
				# handle progress output
				if mode == 1: print "scanning " + root \
					+ " (" + str(len(files)) + " files)"
				elif mode == 2:
					pass

				for name in files:
					if re.search(extRE, string.lower(name)) != None:
						path = os.path.join(root, name)
						try:
							img = Image.open(path)
							if img.getbands() != ('L'):
								img = img.convert('L')
							#imageDB.append(ImageRecord(path,
							#	os.path.getsize(path),
							#	calcImageSig(img, 4)))
							imageDB.append(ImageRecord(
								path[len(rootDir):],
								os.path.getsize(path),
								calcImageSig(img, 4)))
						except IOError:
							print "Unable to open image " + path

	return imageDB


def writeImageDBFile(rootDir, imageDB, outFile):
	"""Write the image DB to a text file."""
	outFile.write(rootDir + "\n\n")
	for rec in imageDB:
		outFile.write(rec.path + "*" + str(rec.fileSize) + " "
			+ string.join(map(str, rec.coefs4x4)) + "\n")


def readImageDBFile(inFile):
	"""
	Reads an *.imagedb file into a list of ImageRecords.
	"""
	rootDir = inFile.readline().rstrip()
	inFile.readline()
	imageDB = []
	for line in inFile.readlines():
		path, rest = string.split(line, '*')
		nums = string.split(rest)
		imageDB.append(ImageRecord(path, int(nums[0]),
			map(float, string.split(rest)[1:])))
	return rootDir, imageDB


def choose2(a):
	"""Returns a generator for enumerating n-choose-2 unordered list pairs."""
	return ((x, y) for i, x in enumerate(a) for j, y in enumerate(a) if i < j)


def genSimilarityList(imageDB, minCorTol = 0.95, mode = 1):
	"""
	Generates a list of similar images from candidates in the image list based
	on 4x4 grid signature correlation.
	"""
	numScans = len(imageDB) - 1
	numPairs = ((numScans + 1) * numScans) / 2
	totalComps = 0

	if mode == 1:
		print "starting scan of %d images..." % (len(imageDB))

	simList = []
	for img1, img2 in choose2(imageDB):
		cor = abs(stats.pearsonr(img1.coefs4x4, img2.coefs4x4)[0])
		if cor > minCorTol:
			 simList.append((cor, img1.path, img2.path))
		totalComps += 1

	if mode == 1:
		print numPairs, "max possible pairs"
		print totalComps, "comparisons made"
		print len(simList), "pairs found above threshold"

	return simList


def writeSimilarityListFile(rootDir, simList, outFile):
	"""Writes a list a matched image pairs to a text file."""
	outFile.write(rootDir + "\n\n")
	for elem in simList:
		outFile.write(str(elem[0]) + "\n" + elem[1] + "\n" + elem[2] + "\n\n")


def fileExtsFromString(fileExtStr = ""):
	"""
	Returns a list of file extensions parsed from the user supplied string.

	Extensions are separated, converted to lowercase, and stripped of all
	separator and wildcard characters.
	"""
	exts = re.split("\W+", string.lower(fileExtStr))
	if len(exts) == 1 and len(exts[0]) == 0:
		exts = []
	return exts


if __name__ == "__main__":
	defExts = "jpg jpeg jpe gif tif tiff png bmp"
	defMinCorTol = 0.95

	numArgs = len(sys.argv)

	# print USAGE and exit if invalid or incorrect number of arguments supplied
	if numArgs < 2 or (sys.argv[1] != "--builddb"
		and sys.argv[1] != "--finddups") or (sys.argv[1] == "--builddb"
		and numArgs < 4) or (sys.argv[1] == "--finddups" and numArgs < 3):
		print "USAGES:"
		print "    SnapMatcher.py --builddb <GROUPNAME> <ROOTDIR> [EXT 1] " \
			"[EXT 2] ..."
		print "    SnapMatcher.py --finddups <GROUPNAME> [TOLERANCE]"
		print "if no extensions specified for --builddb the following " \
			"default extensions will be used:", defExts
		print "if no minimum correlation tolerance specified for --finddups " \
			"then", defMinCorTol, "will be used"
 		sys.exit()

	if sys.argv[1] == "--builddb":
		# convert root search directory to absolute path
		groupName = sys.argv[2]
		rootDir = os.path.abspath(sys.argv[3]) + os.path.sep
		imageDBFilename = groupName + ".imagedb"

		# extract the valid file extensions to check for
		if numArgs > 4:
			fileExts = fileExtsFromString(string.join(sys.argv[4:], " "))
		else: fileExts = string.split(defExts)

		print "Building image database..."
		print "DIR=" + rootDir
		print "EXTS=" + str(fileExts)

		try:
			imageDBFile = open(imageDBFilename, "w")
			imageDB = createImageDB(rootDir, [rootDir], fileExts)
			writeImageDBFile(rootDir, imageDB, imageDBFile)
			imageDBFile.close()
		except IOError:
			print "ERROR: unable to write image database file '" \
				+ imageDBFilename + "'"
			sys.exit()

		print str(len(imageDB)) + " images analyzed"

	elif sys.argv[1] == "--finddups":
		groupName = sys.argv[2]
		imageDBFilename = groupName + ".imagedb"
		if numArgs > 3: minCorTol = float(sys.argv[3])
		else: minCorTol = defMinCorTol

		try:
			imageDBFile = open(imageDBFilename, "r")
			rootDir, imageDB = readImageDBFile(imageDBFile)
			imageDBFile.close()
		except IOError:
			print "ERROR: could not read image database file '" \
				+ imageDBFilename + "'"
			sys.exit()

		resultsFilename = groupName + ".matches"

		try:
			simFile = open(resultsFilename, "w")
			simList = genSimilarityList(imageDB, minCorTol)
			simList.sort(reverse = True)
			writeSimilarityListFile(rootDir, simList, simFile)
			simFile.close()
		except IOError:
			print "ERROR: unable to write file '" + resultsFilename + "'"
			sys.exit()

